//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//
// Program        : Generic Test Driver
//
// Author         : Richard E. Pattis
//                  Computer Science Department
//                  Carnegie Mellon University
//                  5000 Forbes Avenue
//                  Pittsburgh, PA 15213-3891
//                  e-mail: pattis@cs.cmu.edu
//
// Maintainer     : Author
//
//
// Description:
//
//   This amazingly small program uses reflection to test drive any class.
// Just enter the full class name (including its package prefix) and this
// program will construct an object from this class (you choose which
// constructor) and then test all its methods.
//
// Known Bugs     : If a parameter's type is not String or a primitive,
//                    the value null is used (and probably not right)
//
// Program History:
//   10/24/04: R. Pattis - Operational for 15-200
//
//
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


import java.lang.reflect.*;

import edu.cmu.cs.pattis.cs151xx.*;



public class Application {

  //Return a name with the package prefix stripped off
  //packageless(java.lang.String) returns String
  //Package names make members unique, but they frequently
  //  get in the way for simple applications
  static String packageless(String name)
  {return name.substring(name.lastIndexOf(".")+1);}
  
  
  //Return the prototype of a method (including its exceptions)
  //  with all package prefixes stripped off (method name and
  //  parameter types).
  private static String packageless(Constructor c)
  {
    String prototype = " " + c.getName() +"(";
    Class[] parameters = c.getParameterTypes();
    for (int i=0; i<parameters.length; i++)
      prototype += (i==0?"":",")+ packageless(parameters[i].getName());
    prototype += ")";
        
    Class[] exceptions = c.getExceptionTypes();
    for (int i=0; i<exceptions.length; i++)
      prototype += (i==0?"\n       ":",")+ packageless(exceptions[i].getName());
        
    return prototype;
  }
  
  
  //Return the prototype of a method (including its exceptions)
  //  with all package prefixes stripped off (method name and
  //  parameter types).
  private static String packageless(Method m)
  {
    String prototype = packageless(m.getReturnType().toString());
    
    prototype += " " + m.getName() +"(";
    Class[] parameters = m.getParameterTypes();
    for (int i=0; i<parameters.length; i++)
      prototype += (i==0?"":",")+ packageless(parameters[i].getName());
    prototype += ")";
        
    Class[] exceptions = m.getExceptionTypes();
    for (int i=0; i<exceptions.length; i++)
      prototype += (i==0?"\n       ":",")+ packageless(exceptions[i].getName());
        
    return prototype;
  }
  
  
  //Get the appropriate arguments for the parameter classes specified
  //Wrap the primitives as appropriate
  private static Object[] getArguments (Class[] parameters)
  {
    Object[] arguments = new Object[parameters.length];
    
    for (int i=0; i<arguments.length; i++) {
      System.out.print("  Argument #"+(i+1)+": ");
      
      if (parameters[i] == String.class)
        arguments[i] = Prompt.forString("Enter String");

      else if (parameters[i] == int.class)
        arguments[i] = new Integer(Prompt.forInt("Enter int"));

      else if (parameters[i] == double.class)
        arguments[i] = new Double(Prompt.forDouble("Enter double"));

      else if (parameters[i] == char.class)
        arguments[i] = new Character(Prompt.forChar("Enter char"));

      else if (parameters[i] == boolean.class)
        arguments[i] = new Boolean(Prompt.forBoolean("Enter boolean"));
        
      else {
        System.out.println(" Cannot enter parameter of this type; using null");
        arguments[i] = null;
      }  
    }
    return arguments;
  }
  

  
  //Return an object constructed from the specified className
  //This name must be full, prefixed by the package 
  private static Object constructObject(String className) throws ClassNotFoundException
  {
    Constructor[] constructors = Class.forName(className).getConstructors();
    
    for (;;) {
      //Show all constructors and prompt for the one to call
      //Use origin of 1
      System.out.println("Constructors to Call (full prototypes shown)");
      for (int i=0; i<constructors.length; i++)
        System.out.println("  "+(i+1)+") " + packageless(constructors[i]));
      int select = Prompt.forInt("Enter Constructor#",1,constructors.length);
    
      //Select method (just by indexing in methods array) and call it
      //  with appropriate arguments
      //Must at least catch InvocationTargetException and IllegalAccesException
      try {
        Constructor c = constructors[select-1];
        return c.newInstance(getArguments(c.getParameterTypes()));
      } catch (Exception e) {
           System.out.println("  Caught exception: " + e);
           System.out.println("  message = " + e.getMessage());
      }
      return null;
    }
  } 




  //Repeatedly prompt the user for the index of the method to test,
  //  and then for the arguments matching its parameters.
  //Print the returned result (it is null, if the method is void)
  //If the method throws an exception, it will be translated into
  //  an InvocationTargetException, which is caught and reported here.
  private static void testMethods(Object o)
  {
    Method[] methods = o.getClass().getDeclaredMethods();
    
    for (;;) {
      //Print the object's state
      System.out.println("\n\n"+o);
      
      //Show all methods and prompt for the one to call
      //Use origin of 1
      System.out.println("Methods to Call (full prototypes shown); 0 to quit");
      for (int i=0; i<methods.length; i++)
        System.out.println("  "+(i+1)+") " + packageless(methods[i]));
      int select = Prompt.forInt("Enter Method#",0,methods.length);
      if (select == 0)
        return;
        
      //Select method (just by indexing in methods array) and call it
      //  with appropriate arguments
      //Must at least catch InvocationTargetException and IllegalAccesException
      try {
        Method m = methods[select-1];
        Object result = m.invoke(o,getArguments(m.getParameterTypes()));
        System.out.println("  result = " + result);
      } catch (Exception e) {
           System.out.println("  Caught exception: " + e);
           System.out.println("  message = " + e.getMessage());
      }
    }
  } 




  //Just enter the name of a class and it will construct an
  //  object from one of its constructors and then test its methods!
  //F---ing amazing!
	public static void main(String[] args)
	{
	
	  for (;;)
	    try {
        Object o = constructObject(Prompt.forString("\n\nEnter full class (with package) name to test",
                                                    "java.util.StringTokenizer"));
        testMethods(o);
      } catch (ClassNotFoundException cnfe) {System.out.println("Class name not found");}
  }


}
